"""
tl_mono_to_color_processor.py
"""

from ctypes import cdll, POINTER, c_int, c_ushort, c_void_p, c_char_p, c_float, c_ubyte
from typing import Any
from traceback import format_exception
import logging
import platform

import numpy as np

from tl_color_enums import FORMAT, FILTER_ARRAY_PHASE
from tl_camera_enums import SENSOR_TYPE
from tl_mono_to_color_enums import COLOR_SPACE

""" Setup logger """
_logger = logging.getLogger('thorlabs_tsi_sdk.tl_mono_to_color_processor')

""" error-handling methods """


def _get_last_error(sdk):
    try:
        error_pointer = sdk.tl_mono_to_color_get_last_error()
        if error_pointer is None:
            return None
        return str(error_pointer.decode("utf-8"))
    except Exception as exception:
        _logger.error("unable to get last error; " + str(exception))


def _create_c_failure_message(sdk, function_name, error_code):
    last_error = _get_last_error(sdk)
    failure_message = "{function_name}() returned non-zero error code: {error_code}; " \
                      "error message: {last_error}" \
        .format(function_name=function_name, error_code=error_code, last_error=last_error)
    return failure_message


""" Other ctypes types """

# noinspection PyTypeChecker
_3x3Matrix_float = c_float * 9

""" Classes """


class MonoToColorProcessorSDK(object):

    """
    MonoToColorProcessorSDK

    The SDK object that is used to create MonoToColorProcessor objects. There must be only one instance of this class
    active at a time. Use the :meth:`dispose()<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessorSDK.dispose>` method to destroy an SDK instance before
    creating another instance. *with* statements can also be used with this class to automatically dispose the SDK.

    """

    _is_sdk_open = False  # is SDK DLL currently being accessed by a MonoToColorSDK instance

    def __init__(self):
        # type: (type(None)) -> None
        self._disposed = True

        if MonoToColorProcessorSDK._is_sdk_open:
            raise MonoToColorError("MonoToColorProcessorSDK is already in use. Please dispose of the current instance "
                                   "before trying to create another instance.")

        try:
            if platform.system() == 'Windows':
                self._sdk = cdll.LoadLibrary(r"thorlabs_tsi_mono_to_color_processing.dll")
            elif platform.system() == 'Linux':
                self._sdk = cdll.LoadLibrary(r"libthorlabs_tsi_mono_to_color_processing.so")
            else:
                raise MonoToColorError("{system} is not a supported platform.".format(system=platform.system()))
            self._disposed = False
        except OSError as os_error:
            raise MonoToColorError(str(os_error) +
                                   "\nUnable to load library - are the thorlabs tsi mono to color sdk libraries "
                                   "discoverable from the application directory? Try placing them in the same "
                                   "directory as your program, or adding the directory with the libraries to the "
                                   "PATH. Make sure to use 32-bit libraries when using a 32-bit python interpreter "
                                   "and 64-bit libraries when using a 64-bit interpreter.\n")

        error_code = self._sdk.tl_mono_to_color_processing_module_initialize()
        if error_code != 0:
            raise MonoToColorError("tl_mono_to_color_processing_module_initialize() returned error code: {error_code}\n"
                                   .format(error_code=error_code))
        MonoToColorProcessorSDK._is_sdk_open = True

        try:
            """ set C function argument types """
            self._sdk.tl_mono_to_color_create_mono_to_color_processor.argtypes = [c_int, c_int,
                                                                                  POINTER(_3x3Matrix_float),
                                                                                  POINTER(_3x3Matrix_float), c_int,
                                                                                  POINTER(c_void_p)]
            self._sdk.tl_mono_to_color_destroy_mono_to_color_processor.argtypes = [c_void_p]
            self._sdk.tl_mono_to_color_get_color_space.argtypes = [c_void_p, POINTER(c_int)]
            self._sdk.tl_mono_to_color_set_color_space.argtypes = [c_void_p, c_int]
            self._sdk.tl_mono_to_color_get_output_format.argtypes = [c_void_p, POINTER(c_int)]
            self._sdk.tl_mono_to_color_set_output_format.argtypes = [c_void_p, c_int]
            self._sdk.tl_mono_to_color_get_red_gain.argtypes = [c_void_p, POINTER(c_float)]
            self._sdk.tl_mono_to_color_set_red_gain.argtypes = [c_void_p, c_float]
            self._sdk.tl_mono_to_color_get_green_gain.argtypes = [c_void_p, POINTER(c_float)]
            self._sdk.tl_mono_to_color_set_green_gain.argtypes = [c_void_p, c_float]
            self._sdk.tl_mono_to_color_get_blue_gain.argtypes = [c_void_p, POINTER(c_float)]
            self._sdk.tl_mono_to_color_set_blue_gain.argtypes = [c_void_p, c_float]
            self._sdk.tl_mono_to_color_transform_to_48.argtypes = [c_void_p, POINTER(c_ushort), c_int, c_int,
                                                                   POINTER(c_ushort)]
            self._sdk.tl_mono_to_color_transform_to_32.argtypes = [c_void_p, POINTER(c_ushort), c_int, c_int,
                                                                   POINTER(c_ubyte)]
            self._sdk.tl_mono_to_color_transform_to_24.argtypes = [c_void_p, POINTER(c_ushort), c_int, c_int,
                                                                   POINTER(c_ubyte)]
            self._sdk.tl_mono_to_color_processing_module_terminate.argtypes = []
            self._sdk.tl_mono_to_color_get_last_error.restype = c_char_p
            self._sdk.tl_mono_to_color_get_camera_sensor_type.argtypes = [c_void_p, POINTER(c_int)]
            self._sdk.tl_mono_to_color_get_color_filter_array_phase.argtypes = [c_void_p, POINTER(c_int)]
            self._sdk.tl_mono_to_color_get_color_correction_matrix.argtypes = [c_void_p,
                                                                               POINTER(POINTER(_3x3Matrix_float))]
            self._sdk.tl_mono_to_color_get_default_white_balance_matrix.argtypes = [c_void_p,
                                                                                    POINTER(POINTER(_3x3Matrix_float))]
            self._sdk.tl_mono_to_color_get_bit_depth = [c_void_p, POINTER(c_int)]
        except Exception as exception:
            _logger.error("SDK initialization failed; " + str(exception))
            raise exception

    def __del__(self):
        self.dispose()

    """ with statement functionality """

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is not None:
            _logger.debug("".join(format_exception(exception_type, exception_value, exception_traceback)))
        self.dispose()
        return True if exception_type is None else False

    """ methods """

    def dispose(self):
        # type: (type(None)) -> None
        """
        Cleans up the MonoToColorProcessorSDK instance - make sure to call this when you are done with the
        MonoToColorProcessorSDK instance. If using the *with* statement, dispose is called automatically upon exit.

        """
        try:
            if self._disposed:
                return
            error_code = self._sdk.tl_mono_to_color_processing_module_terminate()
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(
                    self._sdk, "tl_mono_to_color_processing_module_terminate", error_code))
            MonoToColorProcessorSDK._is_sdk_open = False
            self._disposed = True
        except Exception as exception:
            _logger.error("Mono To Color SDK destruction failed; " + str(exception))
            raise exception

    def create_mono_to_color_processor(self, camera_sensor_type, color_filter_array_phase, color_correction_matrix,
                                       default_white_balance_matrix, bit_depth):
        # type: (SENSOR_TYPE, FILTER_ARRAY_PHASE, np.array, np.array, int) -> MonoToColorProcessor
        """
        Creates a MonoToColorProcessor object using the given parameters.

        :parameter: - **camera_sensor_type** (:class:`SENSOR_TYPE<thorlabs_tsi_sdk.tl_camera_enums.SENSOR_TYPE>`) - The sensor type used by the camera. Use the property :attr:`TLCamera.camera_sensor_type<thorlabs_tsi_sdk.tl_camera.TLCamera.camera_sensor_type>` to get this information from a camera.
        :parameter: - **color_filter_array_phase** (:class:`FILTER_ARRAY_PHASE<thorlabs_tsi_sdk.tl_color.FILTER_ARRAY_PHASE>`) - The array phase of the camera's color filter. Use :attr:`TLCamera.color_filter_array_phase<thorlabs_tsi_sdk.tl_camera.TLCamera.color_filter_array_phase>` to get this information from a camera.
        :parameter: - **color_correction_matrix** (np.array) - A 3x3 correction matrix specific to a camera model that is used during color processing to achieve accurate coloration. use :meth:`TLCamera.get_color_correction_matrix<thorlabs_tsi_sdk.tl_camera.TLCamera.get_color_correction_matrix>` to get this information from a camera.
        :parameter: - **default_white_balance_matrix** (np.array) - A 3x3 correction matrix specific to a camera model that is used during color processing to white balance images under typical lighting conditions. Use :meth:`TLCamera.get_default_white_balance_matrix<thorlabs_tsi_sdk.tl_camera.TLCamera.get_default_white_balance_matrix>` to get this information from a camera.
        :parameter: - **bit_depth** (int) - The bit depth that will be used during color processing. To get the bit depth of a camera, use :attr:`TLCamera.bit_depth<thorlabs_tsi_sdk.tl_camera.TLCamera.bit_depth>`

        :returns: :class:`MonoToColorProcessor<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor>`

        """
        try:
            c_mono_to_color_handle = c_void_p()
            c_camera_sensor_type = c_int(camera_sensor_type)
            c_color_filter_array_phase = c_int(color_filter_array_phase)
            c_color_correction_matrix = _3x3Matrix_float(*color_correction_matrix)
            c_default_white_balance_matrix = _3x3Matrix_float(*default_white_balance_matrix)
            c_bit_depth = c_int(bit_depth)
            error_code = self._sdk.tl_mono_to_color_create_mono_to_color_processor(c_camera_sensor_type,
                                                                                   c_color_filter_array_phase,
                                                                                   c_color_correction_matrix,
                                                                                   c_default_white_balance_matrix,
                                                                                   c_bit_depth,
                                                                                   c_mono_to_color_handle)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(
                    self._sdk, "tl_mono_to_color_create_mono_to_color_processor", error_code))
            # noinspection PyProtectedMember
            return MonoToColorProcessor._create(self._sdk, c_mono_to_color_handle)
        except Exception as exception:
            _logger.error("Failed to create mono to color processor; " + str(exception))
            raise exception


class MonoToColorProcessor(object):

    """
    MonoToColorProcessor

    These objects are used to quickly convert monochrome image data to colored image data. When finished with a MonoToColorProcessor,
    call its :meth:`dispose<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor.dispose>` method to clean up any opened resources. These object can
    be managed using *with* statements for automatic resource clean up. These objects can only be created by calls to
    :meth:`MonoToColorProcessorSDK.create_mono_to_color_processor()<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessorSDK.create_mono_to_color_processor>`

    """

    __key = object()

    @classmethod
    def _create(cls, sdk, mono_to_color_processor_handle):
        # type: (Any, Any) -> MonoToColorProcessor
        return MonoToColorProcessor(cls.__key, sdk, mono_to_color_processor_handle)

    def __init__(self, key, sdk, mono_to_color_processor_handle):
        # type: (type(object), Any, Any) -> None
        try:
            self._disposed = True
            assert (key == MonoToColorProcessor.__key), \
                "MonoToColorProcessor objects cannot be created manually. Please use " \
                "MonoToColorProcessorSDK.create_mono_to_color_processor to acquire new MonoToColorProcessor objects."
            self._sdk = sdk
            self._mono_to_color_processor_handle = mono_to_color_processor_handle
            self._disposed = False
        except Exception as exception:
            _logger.error("MonoToColorProcessor initialization failed; " + str(exception))
            raise exception

    def __del__(self):
        self.dispose()

    """ with statement functionality """
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is not None:
            _logger.debug("".join(format_exception(exception_type, exception_value, exception_traceback)))
        self.dispose()
        return True if exception_type is None else False

    def dispose(self):
        # type: (type(None)) -> None
        """
        Cleans up the MonoToColorProcessor instance - make sure to call this when you are done with the MonoToColor
        processor. If using the *with* statement, dispose is called automatically upon exit.

        """
        try:
            if self._disposed:
                return
            error_code = self._sdk.tl_mono_to_color_destroy_mono_to_color_processor(
                self._mono_to_color_processor_handle)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(
                    self._sdk, "tl_mono_to_color_destroy_mono_to_color_processor", error_code))
            self._disposed = True
        except Exception as exception:
            _logger.error("Could not dispose MonoToColorProcessor instance; " + str(exception))
            raise exception

    def transform_to_48(self, input_buffer, image_width_pixels, image_height_pixels):
        # type: (np.array, int, int) -> np.array
        """
        Convert monochrome image data into a 3-channel colored image with 16 bits per channel, resulting in 48 bits per
        pixel. The pixel data will be ordered according to the current value of
        :attr:`output_format<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor.output_format>`.

        :param np.array input_buffer: Single channel monochrome image data. The size of this array should be image_width * image_height. The dtype of the array should be ctypes.c_ushort or a type of equivalent size, the image buffer that comes directly from the camera is compatible (see: :meth:`TLCamera.get_pending_frame_or_null()<thorlabs_tsi_sdk.tl_camera.TLCamera.get_pending_frame_or_null>`).
        :param int image_width_pixels: The width of the image in the image_buffer.
        :param int image_height_pixels: The height of the image in the image_buffer.
        :return np.array: 3-channel colored output image, *dtype* = ctypes.c_ushort.
        """
        try:
            output_buffer = np.zeros(shape=(image_width_pixels*image_height_pixels*3,), dtype=c_ushort)
            output_buffer_pointer = output_buffer.ctypes.data_as(POINTER(c_ushort))
            input_buffer_pointer = input_buffer.ctypes.data_as(POINTER(c_ushort))
            c_image_width = c_int(image_width_pixels)
            c_image_height = c_int(image_height_pixels)
            error_code = self._sdk.tl_mono_to_color_transform_to_48(self._mono_to_color_processor_handle,
                                                                    input_buffer_pointer,
                                                                    c_image_width,
                                                                    c_image_height,
                                                                    output_buffer_pointer)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_transform_to_48",
                                                                 error_code))
            return output_buffer
        except Exception as exception:
            _logger.error("Could not transform image to 48bpp; " + str(exception))
            raise exception

    def transform_to_32(self, input_buffer, image_width_pixels, image_height_pixels):
        # type: (np.array, int, int) -> np.array
        """
        Convert monochrome image data into a 4-channel colored image with 8 bits per channel, resulting in 32 bits per
        pixel. The pixel data will be ordered according to the current value of
        :attr:`output_format<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor.output_format>`.

        :param np.array input_buffer: Single channel monochrome image data. The size of this array should be image_width * image_height. The dtype of the array should be ctypes.c_ushort or a type of equivalent size, the image buffer that comes directly from the camera is compatible (see: :meth:`TLCamera.get_pending_frame_or_null()<thorlabs_tsi_sdk.tl_camera.TLCamera.get_pending_frame_or_null>`).
        :param int image_width_pixels: The width of the image in the image_buffer.
        :param int image_height_pixels: The height of the image in the image_buffer.
        :return np.array: 4-channel colored output image, *dtype* = ctypes.c_ubyte.
        """
        try:
            output_buffer = np.zeros(shape=(image_width_pixels*image_height_pixels*4,), dtype=c_ubyte)
            output_buffer_pointer = output_buffer.ctypes.data_as(POINTER(c_ubyte))
            input_buffer_pointer = input_buffer.ctypes.data_as(POINTER(c_ushort))
            c_image_width = c_int(image_width_pixels)
            c_image_height = c_int(image_height_pixels)
            error_code = self._sdk.tl_mono_to_color_transform_to_32(self._mono_to_color_processor_handle,
                                                                    input_buffer_pointer,
                                                                    c_image_width,
                                                                    c_image_height,
                                                                    output_buffer_pointer)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_transform_to_32",
                                                                 error_code))
            return output_buffer
        except Exception as exception:
            _logger.error("Could not transform image to 32bpp; " + str(exception))
            raise exception

    def transform_to_24(self, input_buffer, image_width_pixels, image_height_pixels):
        # type: (np.array, int, int) -> np.array
        """
        Convert monochrome image data into a 3-channel colored image with 8 bits per channel, resulting in 24 bits per
        pixel. The pixel data will be ordered according to the current value of
        :attr:`output_format<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor.output_format>`.

        :param np.array input_buffer: Single channel monochrome image data. The size of this array should be image_width * image_height. The dtype of the array should be ctypes.c_ushort or a type of equivalent size, the image buffer that comes directly from the camera is compatible (see: :meth:`TLCamera.get_pending_frame_or_null()<thorlabs_tsi_sdk.tl_camera.TLCamera.get_pending_frame_or_null>`).
        :param int image_width_pixels: The width of the image in the input_buffer.
        :param int image_height_pixels: The height of the image in the input_buffer.
        :return np.array: 3-channel colored output image, *dtype* = ctypes.c_ubyte.
        """
        try:
            output_buffer = np.zeros(shape=(image_width_pixels*image_height_pixels*3,), dtype=c_ubyte)
            output_buffer_pointer = output_buffer.ctypes.data_as(POINTER(c_ubyte))
            input_buffer_pointer = input_buffer.ctypes.data_as(POINTER(c_ushort))
            c_image_width = c_int(image_width_pixels)
            c_image_height = c_int(image_height_pixels)
            error_code = self._sdk.tl_mono_to_color_transform_to_24(self._mono_to_color_processor_handle,
                                                                    input_buffer_pointer,
                                                                    c_image_width,
                                                                    c_image_height,
                                                                    output_buffer_pointer)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_transform_to_24",
                                                                 error_code))
            return output_buffer
        except Exception as exception:
            _logger.error("Could not transform image to 24bpp; " + str(exception))
            raise exception

    """ Properties """

    @property
    def color_space(self):
        """
        The color space of the mono to color processor. See :class:`COLOR_SPACE<thorlabs_tsi_sdk.tl_mono_to_color_enums.COLOR_SPACE>` for what color spaces
        are available.

        :type: :class:`COLOR_SPACE<thorlabs_tsi_sdk.tl_mono_to_color_enums.COLOR_SPACE>`
        """
        try:
            color_space = c_int()
            error_code = self._sdk.tl_mono_to_color_get_color_space(self._mono_to_color_processor_handle, color_space)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_get_color_space",
                                                                 error_code))
            return COLOR_SPACE(int(color_space.value))
        except Exception as exception:
            _logger.error("Could not get color space; " + str(exception))
            raise exception

    @color_space.setter
    def color_space(self, color_space):
        try:
            c_value = c_int(color_space)
            error_code = self._sdk.tl_mono_to_color_set_color_space(self._mono_to_color_processor_handle, c_value)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_set_color_space",
                                                                 error_code))
        except Exception as exception:
            _logger.error("Could not set color space; " + str(exception))
            raise exception

    @property
    def output_format(self):
        """
        The format of the colored output image. This describes how the data is ordered in the returned buffer from the
        transform functions. By default it is RGB_PIXEL. See :class:`FORMAT<thorlabs_tsi_sdk.tl_color_enums.FORMAT>`.

        :type: :class:`FORMAT<thorlabs_tsi_sdk.tl_mono_to_color_enums.FORMAT>`
        """
        try:
            output_format = c_int()
            error_code = self._sdk.tl_mono_to_color_get_output_format(self._mono_to_color_processor_handle,
                                                                      output_format)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_get_output_format",
                                                                 error_code))
            return FORMAT(int(output_format.value))
        except Exception as exception:
            _logger.error("Could not get output format; " + str(exception))
            raise exception

    @output_format.setter
    def output_format(self, output_format):
        try:
            c_value = c_int(output_format)
            error_code = self._sdk.tl_mono_to_color_set_output_format(self._mono_to_color_processor_handle, c_value)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_set_output_format",
                                                                 error_code))
        except Exception as exception:
            _logger.error("Could not set output format; " + str(exception))
            raise exception

    @property
    def red_gain(self):
        """
        The gain factor that will be applied to the red pixel values in the image. The red intensities will be
        multiplied by this gain value in the final colored image. The default red gain is
        taken from the :attr:`default_white_balance_matrix<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor.default_white_balance_matrix>` that
        is passed in when constructing a
        :meth:`MonoToColorProcessor<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessorSDK.create_mono_to_color_processor>`.

        :type: float
        """
        try:
            red_gain = c_float()
            error_code = self._sdk.tl_mono_to_color_get_red_gain(self._mono_to_color_processor_handle, red_gain)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_get_red_gain",
                                                                 error_code))
            return float(red_gain.value)
        except Exception as exception:
            _logger.error("Could not get red gain value; " + str(exception))
            raise exception

    @red_gain.setter
    def red_gain(self, red_gain):
        try:
            c_value = c_float(red_gain)
            error_code = self._sdk.tl_mono_to_color_set_red_gain(self._mono_to_color_processor_handle, c_value)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_set_red_gain",
                                                                 error_code))
        except Exception as exception:
            _logger.error("Could not set red gain value; " + str(exception))
            raise exception

    @property
    def blue_gain(self):
        """
        The gain factor that will be applied to the red pixel values in the image. The blue intensities will be
        multiplied by this gain value in the final colored image. The default blue gain is
        taken from the :attr:`default_white_balance_matrix<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor.default_white_balance_matrix>` that
        is passed in when constructing a
        :meth:`MonoToColorProcessor<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessorSDK.create_mono_to_color_processor>`.

        :type: float
        """
        try:
            blue_gain = c_float()
            error_code = self._sdk.tl_mono_to_color_get_blue_gain(self._mono_to_color_processor_handle, blue_gain)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_get_blue_gain",
                                                                 error_code))
            return float(blue_gain.value)
        except Exception as exception:
            _logger.error("Could not get blue gain value; " + str(exception))
            raise exception

    @blue_gain.setter
    def blue_gain(self, blue_gain):
        try:
            c_value = c_float(blue_gain)
            error_code = self._sdk.tl_mono_to_color_set_blue_gain(self._mono_to_color_processor_handle, c_value)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_set_blue_gain",
                                                                 error_code))
        except Exception as exception:
            _logger.error("Could not set blue gain value; " + str(exception))
            raise exception

    @property
    def green_gain(self):
        """
        The gain factor that will be applied to the red pixel values in the image. The green intensities will be
        multiplied by this gain value in the final colored image. The default green gain is
        taken from the :attr:`default_white_balance_matrix<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessor.default_white_balance_matrix>` that
        is passed in when constructing a
        :meth:`MonoToColorProcessor<thorlabs_tsi_sdk.tl_mono_to_color_processor.MonoToColorProcessorSDK.create_mono_to_color_processor>`.

        :type: float
        """
        try:
            green_gain = c_float()
            error_code = self._sdk.tl_mono_to_color_get_green_gain(self._mono_to_color_processor_handle, green_gain)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_get_green_gain",
                                                                 error_code))
            return float(green_gain.value)
        except Exception as exception:
            _logger.error("Could not get green gain value; " + str(exception))
            raise exception

    @green_gain.setter
    def green_gain(self, green_gain):
        try:
            c_value = c_float(green_gain)
            error_code = self._sdk.tl_mono_to_color_set_green_gain(self._mono_to_color_processor_handle, c_value)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_set_green_gain",
                                                                 error_code))
        except Exception as exception:
            _logger.error("Could not set green gain value; " + str(exception))
            raise exception

    @property
    def camera_sensor_type(self):
        """
        The sensor type of the camera (monochrome, bayer, etc...). This value is passed in during construction and may
        be read back using this property.

        :type: :class:`SENSOR_TYPE<thorlabs_tsi_sdk.tl_camera_enums.SENSOR_TYPE>`
        """
        try:
            camera_sensor_type = c_int()
            error_code = self._sdk.tl_mono_to_color_get_camera_sensor_type(self._mono_to_color_processor_handle,
                                                                           camera_sensor_type)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_get_camera_sensor_type",
                                                                 error_code))
            return SENSOR_TYPE(int(camera_sensor_type.value))
        except Exception as exception:
            _logger.error("Could not get camera sensor type; " + str(exception))
            raise exception

    @property
    def color_filter_array_phase(self):
        """
        The color filter array phase used in this mono to color processor. This value is passed in during construction
        and may be read back using this property.

        :type: :class:`FILTER_ARRAY_PHASE<thorlabs_tsi_sdk.tl_color.FILTER_ARRAY_PHASE>`
        """
        try:
            color_filter_array_phase = c_int()
            error_code = self._sdk.tl_mono_to_color_get_color_filter_array_phase(self._mono_to_color_processor_handle,
                                                                                 color_filter_array_phase)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(
                    self._sdk, "tl_mono_to_color_get_color_filter_array_phase", error_code))
            return FILTER_ARRAY_PHASE(int(color_filter_array_phase.value))
        except Exception as exception:
            _logger.error("Could not get color filter array phase; " + str(exception))
            raise exception

    @property
    def color_correction_matrix(self):
        """
        The default color correction matrix associated with the mono to color processor. This value is passed in during
        construction and may be read back using this property.

        :type: np.array
        """
        try:
            color_correction_matrix = _3x3Matrix_float()
            error_code = self._sdk.tl_mono_to_color_get_color_correction_matrix(self._mono_to_color_processor_handle,
                                                                                color_correction_matrix)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(
                    self._sdk, "tl_mono_to_color_get_color_correction_matrix", error_code))
            color_correction_matrix_as_np_array = np.array([float(color_correction_matrix[0]),
                                                            float(color_correction_matrix[1]),
                                                            float(color_correction_matrix[2]),
                                                            float(color_correction_matrix[3]),
                                                            float(color_correction_matrix[4]),
                                                            float(color_correction_matrix[5]),
                                                            float(color_correction_matrix[6]),
                                                            float(color_correction_matrix[7]),
                                                            float(color_correction_matrix[8])])
            return color_correction_matrix_as_np_array
        except Exception as exception:
            _logger.error("Could not get color correction matrix; " + str(exception))
            raise exception

    @property
    def default_white_balance_matrix(self):
        """
        The default white balance matrix associated with the mono to color processor. This value is passed in during
        construction and may be read back using this property.

        :type: np.array
        """
        try:
            default_white_balance_matrix = _3x3Matrix_float()
            error_code = self._sdk.tl_mono_to_color_get_default_white_balance_matrix(
                self._mono_to_color_processor_handle, default_white_balance_matrix)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(
                    self._sdk, "tl_mono_to_color_get_default_white_balance_matrix", error_code))
            default_white_balance_matrix_as_np_array = np.array([float(default_white_balance_matrix[0]),
                                                                float(default_white_balance_matrix[1]),
                                                                float(default_white_balance_matrix[2]),
                                                                float(default_white_balance_matrix[3]),
                                                                float(default_white_balance_matrix[4]),
                                                                float(default_white_balance_matrix[5]),
                                                                float(default_white_balance_matrix[6]),
                                                                float(default_white_balance_matrix[7]),
                                                                float(default_white_balance_matrix[8])])
            return default_white_balance_matrix_as_np_array
        except Exception as exception:
            _logger.error("Could not get default white balance matrix; " + str(exception))
            raise exception

    @property
    def bit_depth(self):
        """
        The bit depth associated with the mono to color processor. This value is passed in during construction and may
        be read back using this property.

        :type: int
        """
        try:
            bit_depth = c_int()
            error_code = self._sdk.tl_mono_to_color_get_bit_depth(self._mono_to_color_processor_handle, bit_depth)
            if error_code != 0:
                raise MonoToColorError(_create_c_failure_message(self._sdk, "tl_mono_to_color_get_bit_depth",
                                                                 error_code))
            return int(bit_depth.value)
        except Exception as exception:
            _logger.error("Could not get bit depth; " + str(exception))
            raise exception


""" Error handling """


class MonoToColorError(Exception):
    def __init__(self, message):
        _logger.debug(message)
        super(MonoToColorError, self).__init__(message)



# -*- coding: utf-8 -*-
"""
Kent-CAS: Camera Acquisition System

Test Kiralux Camera

@author: Mike Hughes
Applied Optics Group, University of Kent
"""

import context

import matplotlib.pyplot as plt

from KiraluxCamera import KiraluxCamera
import sys

sys.path.append(r"\\dll64_lib")

cam = KiraluxCamera()

cam.open_camera(0)

im = cam.get_image()
im2 = cam.get_image()

plt.figure()
plt.imshow(im)

cam.close_camera()






"""
tl_mono_to_color_processor.py ***BETA***
"""

from ctypes import cdll, POINTER, c_int, c_ushort, c_void_p, c_char_p, c_float, c_ubyte
from typing import Any
from traceback import format_exception
import logging
import platform

import numpy as np

from .tl_polarization_enums import POLAR_PHASE

""" Setup logger """
logging.basicConfig(level=logging.ERROR,
                    format='%(module)s %(asctime)s %(levelname)s: %(message)s')
_logger = logging.getLogger('thorlabs_tsi_sdk.tl_polarization_processor')

""" error-handling methods """


def _get_last_error(sdk):
    return ""  # tl_polarization_processor_get_last_error()


def _create_c_failure_message(sdk, function_name, error_code):
    last_error = _get_last_error(sdk)
    failure_message = "{function_name}() returned non-zero error code: {error_code}; " \
                      "error message: {last_error}" \
        .format(function_name=function_name, error_code=error_code, last_error=last_error)
    return failure_message


""" Other ctypes types """

_4x4Matrix_float = c_float * 16

""" Classes """


class PolarizationProcessorSDK(object):

    """
    PolarizationProcessorSDK

    The polarization processor SDK loads DLLs into memory and provides functions to polarization processor instances.
    Be sure to dispose all PolarizationProcessor objects and then dispose the PolarizationProcessorSDK before exiting
    the application. *with* statements can also be used with this class to automatically dispose the SDK.

    """

    _is_sdk_open = False  # is SDK DLL currently being accessed by a MonoToColorSDK instance

    def __init__(self):
        # type: (type(None)) -> None
        self._disposed = True

        if PolarizationProcessorSDK._is_sdk_open:
            raise PolarizationError("PolarizationProcessorSDK is already in use. Please dispose of the current "
                                    "instance before trying to create another instance.")

        try:
            if platform.system() == 'Windows':
                self._sdk = cdll.LoadLibrary(r"thorlabs_tsi_polarization_processor.dll")
            elif platform.system() == 'Linux':
                self._sdk = cdll.LoadLibrary(r"libthorlabs_tsi_polarization_processor.so")
            else:
                raise PolarizationError("{system} is not a supported platform.".format(system=platform.system()))
            self._disposed = False
        except OSError as os_error:
            raise PolarizationError(str(os_error) +
                                    "\nUnable to load library - are the thorlabs tsi polarization libraries "
                                    "discoverable from the application directory? Try placing them in the same "
                                    "directory as your program, or adding the directory with the libraries to the "
                                    "PATH. Make sure to use 32-bit libraries when using a 32-bit python interpreter "
                                    "and 64-bit libraries when using a 64-bit interpreter.\n")

        error_code = self._sdk.tl_polarization_processor_module_initialize()
        if error_code != 0:
            raise PolarizationError("tl_polarization_processing_module_initialize() returned error code: {error_code}\n"
                                    .format(error_code=error_code))
        PolarizationProcessorSDK._is_sdk_open = True

        try:
            """ set C function argument types """
            self._sdk.tl_polarization_processor_create_polarization_processor.argtypes = [POINTER(c_void_p)]
            self._sdk.tl_polarization_processor_destroy_polarization_processor.argtypes = [c_void_p]
            self._sdk.tl_polarization_processor_set_custom_calibration_coefficients.argtypes = \
                [c_void_p,
                 POINTER(_4x4Matrix_float),
                 POINTER(_4x4Matrix_float),
                 POINTER(_4x4Matrix_float),
                 POINTER(_4x4Matrix_float)]
            self._sdk.tl_polarization_processor_get_custom_calibration_coefficients.argtypes = \
                [c_void_p,
                 POINTER(_4x4Matrix_float),
                 POINTER(_4x4Matrix_float),
                 POINTER(_4x4Matrix_float),
                 POINTER(_4x4Matrix_float)]
            self._sdk.tl_polarization_processor_transform.argtypes = [c_void_p,
                                                                      c_int,
                                                                      POINTER(c_ushort),
                                                                      c_int,
                                                                      c_int,
                                                                      c_int,
                                                                      c_int,
                                                                      c_int,
                                                                      c_ushort,
                                                                      POINTER(c_float),
                                                                      POINTER(c_ushort),
                                                                      POINTER(c_ushort),
                                                                      POINTER(c_ushort),
                                                                      POINTER(c_ushort),
                                                                      POINTER(c_ushort),
                                                                      ]
            self._sdk.tl_polarization_processor_destroy_polarization_processor.argtypes = [c_void_p]
        except Exception as exception:
            _logger.error("SDK initialization failed; " + str(exception))
            raise exception

    def __del__(self):
        self.dispose()

    """ with statement functionality """

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is not None:
            _logger.debug("".join(format_exception(exception_type, exception_value, exception_traceback)))
        self.dispose()
        return True if exception_type is None else False

    """ methods """

    def dispose(self):
        # type: (type(None)) -> None
        """
        Cleans up the PolarizationProcessorSDK instance - make sure to call this when you are done with the
        PolarizationProcessorSDK instance. If using the *with* statement, dispose is called automatically upon exit.

        """
        try:
            if self._disposed:
                return
            error_code = self._sdk.tl_polarization_processor_module_terminate()
            if error_code != 0:
                raise PolarizationError(_create_c_failure_message(
                    self._sdk, "tl_polarization_processor_module_terminate", error_code))
            PolarizationProcessorSDK._is_sdk_open = False
            self._disposed = True
        except Exception as exception:
            _logger.error("Polarization SDK destruction failed; " + str(exception))
            raise exception

    def create_polarization_processor(self):
        # type: (type(None)) -> PolarizationProcessor
        """
        Creates a Polarization Processor object.

        :returns: :class:`PolarizationProcessor<thorlabs_tsi_sdk.tl_polarization_processor.PolarizationProcessor>`

        """
        try:
            c_polarization_handle = c_void_p()
            error_code = self._sdk.tl_polarization_processor_create_polarization_processor(c_polarization_handle)
            if error_code != 0:
                raise PolarizationError(_create_c_failure_message(
                    self._sdk, "tl_polarization_processor_create_polarization_processor", error_code))
            # noinspection PyProtectedMember
            return PolarizationProcessor._create(self._sdk, c_polarization_handle)
        except Exception as exception:
            _logger.error("Failed to create polarization processor; " + str(exception))
            raise exception


class PolarizationProcessor(object):

    """
    PolarizationProcessor

    These objects are used to convert raw sensor data to polarized data. When finished with a PolarizationProcessor,
    call its :meth:`dispose<thorlabs_tsi_sdk.tl_polarization_processor.PolarizationProcessor.dispose>` method to clean up any opened resources. These object can
    be managed using *with* statements for automatic resource clean up. These objects can only be created by calls to
    :meth:`PolarizationProcessorSDK.create_polarization_processor()<thorlabs_tsi_sdk.tl_polarization_processor.PolarizationProcessorSDK.create_polarization_processor>`

    The transform functions return an output image with values from 0 to max value for each pixel. To calculate
    scaled values for each pixel, you may use the following equation:

    • polarization_parameter_value = pixel_integer_value * (max_parameter_value / (2^bit_depth - 1)) + min_parameter_value

    The suggested min and max for each type are as follows:

    • Azimuth: (range from -90° to 90°)
        ◦ min_parameter_value = -90
        ◦ max_parameter_value = 180

    • DoLP: (range from 0 to 100%)
        ◦ min_parameter_value = 0
        ◦ max_parameter_value = 100

    • Intensity: (range from 0 to 100)
        ◦ min_parameter_value = 0
        ◦ max_parameter_value = 100

    """

    __key = object()

    @classmethod
    def _create(cls, sdk, polarization_processor_handle):
        # type: (Any, Any) -> PolarizationProcessor
        return PolarizationProcessor(cls.__key, sdk, polarization_processor_handle)

    def __init__(self, key, sdk, polarization_processor_handle):
        # type: (type(object), Any, Any) -> None
        try:
            self._disposed = True
            assert (key == PolarizationProcessor.__key), \
                "PolarizationProcessor objects cannot be created manually. Please use " \
                "PolarizationProcessorSDK.create_polarization_processor to acquire new PolarizationProcessor objects."
            self._sdk = sdk
            self._polarization_processor_handle = polarization_processor_handle
            self._disposed = False
        except Exception as exception:
            _logger.error("PolarizationProcessor initialization failed; " + str(exception))
            raise exception

    def __del__(self):
        self.dispose()

    """ with statement functionality """
    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is not None:
            _logger.debug("".join(format_exception(exception_type, exception_value, exception_traceback)))
        self.dispose()
        return True if exception_type is None else False

    def dispose(self):
        # type: (type(None)) -> None
        """
        Cleans up the PolarizationProcessor instance - make sure to call this when you are done with the polarization
        processor. If using the *with* statement, dispose is called automatically upon exit.

        """
        try:
            if self._disposed:
                return
            error_code = self._sdk.tl_polarization_processor_destroy_polarization_processor(
                self._polarization_processor_handle)
            if error_code != 0:
                raise PolarizationError(_create_c_failure_message(
                    self._sdk, "tl_polarization_processor_destroy_polarization_processor", error_code))
            self._disposed = True
        except Exception as exception:
            _logger.error("Could not dispose PolarizationProcessor instance; " + str(exception))
            raise exception

    def transform_to_intensity(self,
                               sensor_polar_phase,
                               input_image,
                               image_origin_x_pixels,
                               image_origin_y_pixels,
                               image_width_pixels,
                               image_height_pixels,
                               input_image_bit_depth,
                               output_max_value):
        # type: (POLAR_PHASE, np.array, int, int, int, int, int, int) -> np.array
        """
        Transforms raw-image data into the intensity-output buffer, which shows the brightness of the light at each pixel.

        :param sensor_polar_phase: The polar phase (in degrees) of the origin (top-left) pixel of the camera sensor.
        :param input_image: Unprocessed input image delivered by the camera.
        :param image_origin_x_pixels: The X position of the origin (top-left) of the input image on the sensor. Warning: Some camera models may nudge input ROI values. Please read values back from camera. See :attr:`TLCamera.roi<thorlabs_tsi_sdk.tl_camera.TLCamera.roi>`.
        :param image_origin_y_pixels: The Y position of the origin (top-left) of the input image on the sensor. Warning: Some camera models may nudge input ROI values. Please read values back from camera. See :attr:`TLCamera.roi<thorlabs_tsi_sdk.tl_camera.TLCamera.roi>`.
        :param image_width_pixels: Width of input image in pixels
        :param image_height_pixels: Height of input image in pixels
        :param input_image_bit_depth: Bit depth of input image pixels
        :param output_max_value: The maximum possible pixel value in the output images. Must be between 1 and 65535.
        :return np.array: Array containing the total optical power (intensity) output, *dtype* = ctypes.c_ushort.
        """
        try:
            output_buffer = np.zeros(shape=(image_width_pixels * image_height_pixels,), dtype=c_ushort)
            output_buffer_pointer = output_buffer.ctypes.data_as(POINTER(c_ushort))
            input_buffer_pointer = input_image.ctypes.data_as(POINTER(c_ushort))
            c_image_width = c_int(image_width_pixels)
            c_image_height = c_int(image_height_pixels)
            c_image_origin_x_pixels = c_int(image_origin_x_pixels)
            c_image_origin_y_pixels = c_int(image_origin_y_pixels)
            c_input_image_bit_depth = c_int(input_image_bit_depth)
            c_output_max_value = c_ushort(output_max_value)
            c_sensor_polar_phase = c_int(sensor_polar_phase)
            error_code = self._sdk.tl_polarization_processor_transform(self._polarization_processor_handle,
                                                                       c_sensor_polar_phase,
                                                                       input_buffer_pointer,
                                                                       c_image_origin_x_pixels,
                                                                       c_image_origin_y_pixels,
                                                                       c_image_width,
                                                                       c_image_height,
                                                                       c_input_image_bit_depth,
                                                                       c_output_max_value,
                                                                       None,  # normalized_stokes_vector_coefficients_x2
                                                                       output_buffer_pointer,  # total_optical_power
                                                                       None,  # horizontal_vertical_linear_polarization
                                                                       None,  # diagonal_linear_polarization
                                                                       None,  # azimuth
                                                                       None)  # DOLP
            if error_code != 0:
                raise PolarizationError(_create_c_failure_message(self._sdk, "tl_polarization_processor_transform",
                                                                  error_code))
            return output_buffer
        except Exception as exception:
            _logger.error("Could not transform to intensity; " + str(exception))
            raise exception

    def transform_to_dolp(self,
                          sensor_polar_phase,
                          input_image,
                          image_origin_x_pixels,
                          image_origin_y_pixels,
                          image_width_pixels,
                          image_height_pixels,
                          input_image_bit_depth,
                          output_max_value):
        # type: (POLAR_PHASE, np.array, int, int, int, int, int, int) -> np.array
        """
        Transforms raw-image data into a DoLP (degree of linear polarization) output buffer, which is a measure of how polarized the light is from none to totally polarized.

        :param sensor_polar_phase: The polar phase (in degrees) of the origin (top-left) pixel of the camera sensor.
        :param input_image: Unprocessed input image delivered by the camera.
        :param image_origin_x_pixels: The X position of the origin (top-left) of the input image on the sensor. Warning: Some camera models may nudge input ROI values. Please read values back from camera. See :attr:`TLCamera.roi<thorlabs_tsi_sdk.tl_camera.TLCamera.roi>`.
        :param image_origin_y_pixels: The Y position of the origin (top-left) of the input image on the sensor. Warning: Some camera models may nudge input ROI values. Please read values back from camera. See :attr:`TLCamera.roi<thorlabs_tsi_sdk.tl_camera.TLCamera.roi>`.
        :param image_width_pixels: Width of input image in pixels
        :param image_height_pixels: Height of input image in pixels
        :param input_image_bit_depth: Bit depth of input image pixels
        :param output_max_value: The maximum possible pixel value in the output images. Must be between 1 and 65535.
        :return np.array: Array containing the DoLP (degree of linear polarization) output, *dtype* = ctypes.c_ushort.
        """
        try:
            output_buffer = np.zeros(shape=(image_width_pixels * image_height_pixels,), dtype=c_ushort)
            output_buffer_pointer = output_buffer.ctypes.data_as(POINTER(c_ushort))
            input_buffer_pointer = input_image.ctypes.data_as(POINTER(c_ushort))
            c_image_width = c_int(image_width_pixels)
            c_image_height = c_int(image_height_pixels)
            c_image_origin_x_pixels = c_int(image_origin_x_pixels)
            c_image_origin_y_pixels = c_int(image_origin_y_pixels)
            c_input_image_bit_depth = c_int(input_image_bit_depth)
            c_output_max_value = c_ushort(output_max_value)
            c_sensor_polar_phase = c_int(sensor_polar_phase)
            error_code = self._sdk.tl_polarization_processor_transform(self._polarization_processor_handle,
                                                                       c_sensor_polar_phase,
                                                                       input_buffer_pointer,
                                                                       c_image_origin_x_pixels,
                                                                       c_image_origin_y_pixels,
                                                                       c_image_width,
                                                                       c_image_height,
                                                                       c_input_image_bit_depth,
                                                                       c_output_max_value,
                                                                       None,  # normalized_stokes_vector_coefficients_x2
                                                                       None,  # total_optical_power
                                                                       None,  # horizontal_vertical_linear_polarization
                                                                       None,  # diagonal_linear_polarization
                                                                       None,  # azimuth
                                                                       output_buffer_pointer)  # DoLP
            if error_code != 0:
                raise PolarizationError(_create_c_failure_message(self._sdk, "tl_polarization_processor_transform",
                                                                  error_code))
            return output_buffer
        except Exception as exception:
            _logger.error("Could not transform to DoLP; " + str(exception))
            raise exception

    def transform_to_azimuth(self,
                             sensor_polar_phase,
                             input_image,
                             image_origin_x_pixels,
                             image_origin_y_pixels,
                             image_width_pixels,
                             image_height_pixels,
                             input_image_bit_depth,
                             output_max_value):
        # type: (POLAR_PHASE, np.array, int, int, int, int, int, int) -> np.array
        """
        Transforms raw-image data into an azimuth-output buffer, which shows the angle of polarized light at each pixel.

        :param sensor_polar_phase: The polar phase (in degrees) of the origin (top-left) pixel of the camera sensor.
        :param input_image: Unprocessed input image delivered by the camera.
        :param image_origin_x_pixels: The X position of the origin (top-left) of the input image on the sensor. Warning: Some camera models may nudge input ROI values. Please read values back from camera. See :attr:`TLCamera.roi<thorlabs_tsi_sdk.tl_camera.TLCamera.roi>`.
        :param image_origin_y_pixels: The Y position of the origin (top-left) of the input image on the sensor. Warning: Some camera models may nudge input ROI values. Please read values back from camera. See :attr:`TLCamera.roi<thorlabs_tsi_sdk.tl_camera.TLCamera.roi>`.
        :param image_width_pixels: Width of input image in pixels
        :param image_height_pixels: Height of input image in pixels
        :param input_image_bit_depth: Bit depth of input image pixels
        :param output_max_value: The maximum possible pixel value in the output images. Must be between 1 and 65535.
        :return np.array: Array containing the azimuth (polar angle) output, *dtype* = ctypes.c_ushort.
        """
        try:
            output_buffer = np.zeros(shape=(image_width_pixels * image_height_pixels,), dtype=c_ushort)
            output_buffer_pointer = output_buffer.ctypes.data_as(POINTER(c_ushort))
            input_buffer_pointer = input_image.ctypes.data_as(POINTER(c_ushort))
            c_image_width = c_int(image_width_pixels)
            c_image_height = c_int(image_height_pixels)
            c_image_origin_x_pixels = c_int(image_origin_x_pixels)
            c_image_origin_y_pixels = c_int(image_origin_y_pixels)
            c_input_image_bit_depth = c_int(input_image_bit_depth)
            c_output_max_value = c_ushort(output_max_value)
            c_sensor_polar_phase = c_int(sensor_polar_phase)
            error_code = self._sdk.tl_polarization_processor_transform(self._polarization_processor_handle,
                                                                       c_sensor_polar_phase,
                                                                       input_buffer_pointer,
                                                                       c_image_origin_x_pixels,
                                                                       c_image_origin_y_pixels,
                                                                       c_image_width,
                                                                       c_image_height,
                                                                       c_input_image_bit_depth,
                                                                       c_output_max_value,
                                                                       None,  # normalized_stokes_vector_coefficients_x2
                                                                       None,  # total_optical_power
                                                                       None,  # horizontal_vertical_linear_polarization
                                                                       None,  # diagonal_linear_polarization
                                                                       output_buffer_pointer,  # azimuth
                                                                       None)  # DoLP
            if error_code != 0:
                raise PolarizationError(_create_c_failure_message(self._sdk, "tl_polarization_processor_transform",
                                                                  error_code))
            return output_buffer
        except Exception as exception:
            _logger.error("Could not transform to azimuth; " + str(exception))
            raise exception

    """ Properties """


""" Error handling """


class PolarizationError(Exception):
    def __init__(self, message):
        _logger.debug(message)
        super(PolarizationError, self).__init__(message)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul  5 16:27:13 2022

@author: AOG
"""



import cv2 as cv
#import pybundle
import numpy as np
import time
import sys
import os
sys.path.append(os.path.abspath("..\\src\\widgets"))
sys.path.append(os.path.abspath("..\\src\\cameras"))
sys.path.append(os.path.abspath("..\\src\\threads"))

from BundleProcessor import BundleProcessor


imageProcessor = BundleProcessor(10,10, mosaic = True)
imageProcessor.start()
imageProcessor.pyb.set_core_method(imageProcessor.pyb.FILTER)
imageProcessor.pyb.set_filter_size(1.4)
imageProcessor.pyb.set_bundle_loc((600,500,450))
imageProcessor.pyb.set_auto_mask(None)

# Load in a fibre bundle endomicroscopy video
cap = cv.VideoCapture(r"C:\Users\\AOG\\Dropbox\\Programming\\Python\\pybundle\\test\\data\\raw_example.avi")
ret, img = cap.read()
img = img[:,:,0]
nFrames = int(cap.get(cv.CAP_PROP_FRAME_COUNT))


# Load in the calibration image
calibImg = cv.imread(r"C:\Users\\AOG\\Dropbox\\Programming\\Python\\pybundle\\test\\data\\raw_example_calib.tif")[:,:,0]

ret, img = cap.read()
img = img[:,:,0]
imgStack = np.zeros([nFrames, np.shape(img)[0], np.shape(img)[1] ], dtype='uint8'  ) 
imgStack[0,:,:] = img

# Load video frames
for i in range(1,nFrames):
    cap.set(cv.CAP_PROP_POS_FRAMES, i)
    ret, img = cap.read()
    img = img[:,:,0]
    imgStack[i,:,:] = img
print("Data Loaded")


for i in range(1,nFrames):
    print("Processed frame:" + str(i))
    print(imageProcessor.is_image_ready())
    imageProcessor.add_image(imgStack[i,:,:])
    time.sleep(0.01)
    currentProcessedImage = imageProcessor.get_latest_processed_image()
    
    
imageProcessor.stop() 



"""
windows_setup.py

In order for the Thorlabs Python examples to work, they need visibility of the directory containing the Thorlabs TSI
Native DLLs. This setup function changes the PATH environment variable (Just for the current process, not the system
PATH variable) by adding the directory containing the DLLs. This function is written specifically to work for the
Thorlabs Python SDK examples on Windows, but can be adjusted to work with custom programs. Changing the PATH variable
of a running application is just one way of making the DLLs visible to the program. The following methods could
be used instead:

- Use the os module to adjust the program's current directory to be the directory containing the DLLs.
- Manually copy the DLLs into the working directory of your application.
- Manually add the path to the directory containing the DLLs to the system PATH environment variable.

"""

import os
import sys


def configure_path():
    is_64bits = sys.maxsize > 2**32
    relative_path_to_dlls = '..' + os.sep + 'thorlab' + os.sep

  #  if is_64bits:
   #     relative_path_to_dlls += '64_lib'
    #else:
     #   relative_path_to_dlls += '32_lib'

    absolute_path_to_file_directory = os.path.dirname(os.path.abspath(__file__))

    absolute_path_to_dlls = os.path.abspath(absolute_path_to_file_directory + os.sep + relative_path_to_dlls)

    os.environ['PATH'] = absolute_path_to_dlls + os.pathsep + os.environ['PATH']

    try:
        # Python 3.8 introduces a new method to specify dll directory
        os.add_dll_directory(absolute_path_to_dlls)
    except AttributeError:
        pass



# -*- coding: utf-8 -*-
"""
Kent-CAS: Camera Acquisition System

Test Simulated Camera

@author: Mike Hughes
Applied Optics Group, University of Kent
"""

import context

import matplotlib.pyplot as plt
import time

from SimulatedCamera import SimulatedCamera
from DCXCameraInterface import DCXCameraInterface


filename = "C:\\Users\\AOG\\OneDrive - University of Kent\\Experimental\\Holography\\Off Axis Holography\\Example Data\\01_11_21 example OAH data USAF HeNe Kiralux\\background.tif"
filename = r"test_data\stack_10.tif"


# Test one frame at a time
cam = DCXCameraInterface()

cam.open_camera(0)
#cam.pre_load(10)
cam.set_frame_rate_on()
#cam.disable_frame_rate()   # So we see the actual time to load the image

t1 = time.time()
im = cam.get_image()

print("Single frame acquisition took " + str(time.time() - t1 ))

plt.figure()
plt.imshow(im)

cam.close_camera()




"""
tl_mono_to_color_enums.py
"""

from enum import IntEnum


class _CTypesEnum(IntEnum):
    @classmethod
    def from_param(cls, obj):
        return int(obj)


class COLOR_SPACE(_CTypesEnum):
    """
    A color space describes how the colors in an image are going to be specified. Some commonly used color spaces are those derived from the
    RGB color model, in which each pixel has a Red, Blue, and Green component. This means the amount of color that can expressed in a single
    pixel is all the possible combinations of Red, Blue, and Green. If we assume the image data is in bytes, each component can take any value
    from 0 to 255. The total number of colors that a pixel could express can be calculated as 256 * 256 * 256 = 16777216 different colors.

    There are many different color spaces that are used for different purposes. The mono to color processor supports two color spaces that
    are both derived from the RGB color model: sRGB and Linear sRGB.

    """
    SRGB = 0
    """
    sRGB or standard RGB is a common color space used for displaying images on computer monitors or for sending images over the internet. 
    In addition to the Red, Blue, and Green components combining to define the color of a pixel, the final RGB values undergo a nonlinear transformation 
    to be put in the sRGB color space. The exact transfer function can be found online by searching for the sRGB specification. The purpose of this 
    transformation is to represent the colors in a way that looks more accurate to humans.
    
    """
    LINEAR_SRGB = 1
    """
    Linear sRGB is very similar to sRGB, but does not perform the non linear transformation. The transformation of the data in sRGB changes the RGB intensities, 
    whereas this color space is much more representative of the raw image data coming off the sensor. Without the transformation, however, images in the Linear 
    sRGB color space do not look as accurate as those in sRGB. When deciding between Linear sRGB and sRGB, use Linear sRGB when the actual intensities of the raw 
    image data are important and use sRGB when the image needs to look accurate to the human eye.
    
    """



"""
Camera Live View - TkInter

This example shows how one could create a live image viewer using TkInter.
It also uses the third party library 'pillow', which is a fork of PIL.

This example detects if a camera is a color camera and will process the
images using the tl_mono_to_color_processor module.

This example uses threading to enqueue images coming off the camera in one thread, and
dequeue them in the UI thread for quick displaying.

"""

# try:
#     # if on Windows, use the provided setup script to add the DLLs folder to the PATH
#     from windows_setup import configure_path
#     configure_path()
# except ImportError:
#     configure_path = None

from tl_camera import TLCameraSDK, TLCamera, Frame
from tl_camera_enums import SENSOR_TYPE
from tl_mono_to_color_processor import MonoToColorProcessorSDK

try:
    #  For python 2.7 tkinter is named Tkinter
    import Tkinter as tk
except ImportError:
    import tkinter as tk
from PIL import Image, ImageTk
import typing
import threading
try:
    #  For Python 2.7 queue is named Queue
    import Queue as queue
except ImportError:
    import queue

""" LiveViewCanvas

This is a Tkinter Canvas object that can be reused in custom programs. The Canvas expects a parent Tkinter object and 
an image queue. The image queue is a queue.Queue that it will pull images from, and is expected to hold PIL Image 
objects that will be displayed to the canvas. It automatically adjusts its size based on the incoming image dimensions.

"""


class LiveViewCanvas(tk.Canvas):

    def __init__(self, parent, image_queue):
        # type: (typing.Any, queue.Queue) -> LiveViewCanvas
        self.image_queue = image_queue
        self._image_width = 0
        self._image_height = 0
        tk.Canvas.__init__(self, parent)
        self.pack()
        self._get_image()

    def _get_image(self):
        try:
            image = self.image_queue.get_nowait()
            self._image = ImageTk.PhotoImage(master=self, image=image)
            if (self._image.width() != self._image_width) or (self._image.height() != self._image_height):
                # resize the canvas to match the new image size
                self._image_width = self._image.width()
                self._image_height = self._image.height()
                self.config(width=self._image_width, height=self._image_height)
            self.create_image(0, 0, image=self._image, anchor='nw')
        except queue.Empty:
            pass
        self.after(10, self._get_image)


""" ImageAcquisitionThread

This class derives from threading.Thread and is given a TLCamera instance during initialization. When started, the 
thread continuously acquires frames from the camera and converts them to PIL Image objects. These are placed in a 
queue.Queue object that can be retrieved using get_output_queue(). The thread doesn't do any arming or triggering, 
so users will still need to setup and control the camera from a different thread. Be sure to call stop() when it is 
time for the thread to stop.

"""


class ImageAcquisitionThread(threading.Thread):

    def __init__(self, camera):
        # type: (TLCamera) -> ImageAcquisitionThread
        super(ImageAcquisitionThread, self).__init__()
        self._camera = camera
        self._previous_timestamp = 0

        # setup color processing if necessary
        if self._camera.camera_sensor_type != SENSOR_TYPE.BAYER:
            # Sensor type is not compatible with the color processing library
            self._is_color = False
        else:
            self._mono_to_color_sdk = MonoToColorProcessorSDK()
            self._image_width = self._camera.image_width_pixels
            self._image_height = self._camera.image_height_pixels
            self._mono_to_color_processor = self._mono_to_color_sdk.create_mono_to_color_processor(
                SENSOR_TYPE.BAYER,
                self._camera.color_filter_array_phase,
                self._camera.get_color_correction_matrix(),
                self._camera.get_default_white_balance_matrix(),
                self._camera.bit_depth
            )
            self._is_color = True

        self._bit_depth = camera.bit_depth
        self._camera.image_poll_timeout_ms = 0  # Do not want to block for long periods of time
        self._image_queue = queue.Queue(maxsize=2)
        self._stop_event = threading.Event()

    def get_output_queue(self):
        # type: (type(None)) -> queue.Queue
        return self._image_queue

    def stop(self):
        self._stop_event.set()

    def _get_color_image(self, frame):
        # type: (Frame) -> Image
        # verify the image size
        width = frame.image_buffer.shape[1]
        height = frame.image_buffer.shape[0]
        if (width != self._image_width) or (height != self._image_height):
            self._image_width = width
            self._image_height = height
            print("Image dimension change detected, image acquisition thread was updated")
        # color the image. transform_to_24 will scale to 8 bits per channel
        color_image_data = self._mono_to_color_processor.transform_to_24(frame.image_buffer,
                                                                         self._image_width,
                                                                         self._image_height)
        color_image_data = color_image_data.reshape(self._image_height, self._image_width, 3)
        # return PIL Image object
        return Image.fromarray(color_image_data, mode='RGB')

    def _get_image(self, frame):
        # type: (Frame) -> Image
        # no coloring, just scale down image to 8 bpp and place into PIL Image object
        scaled_image = frame.image_buffer >> (self._bit_depth - 8)
        return Image.fromarray(scaled_image)

    def run(self):
        while not self._stop_event.is_set():
            try:
                frame = self._camera.get_pending_frame_or_null()
                if frame is not None:
                    if self._is_color:
                        pil_image = self._get_color_image(frame)
                    else:
                        pil_image = self._get_image(frame)
                    self._image_queue.put_nowait(pil_image)
            except queue.Full:
                # No point in keeping this image around when the queue is full, let's skip to the next one
                pass
            except Exception as error:
                print("Encountered error: {error}, image acquisition will stop.".format(error=error))
                break
        print("Image acquisition has stopped")
        if self._is_color:
            self._mono_to_color_processor.dispose()
            self._mono_to_color_sdk.dispose()


""" Main

When run as a script, a simple Tkinter app is created with just a LiveViewCanvas widget. 

"""
if __name__ == "__main__":
    with TLCameraSDK() as sdk:
        camera_list = sdk.discover_available_cameras()
        with sdk.open_camera(camera_list[0]) as camera:

            # create generic Tk App with just a LiveViewCanvas widget
            print("Generating app...")
            root = tk.Tk()
            root.title(camera.name)
            image_acquisition_thread = ImageAcquisitionThread(camera)
            camera_widget = LiveViewCanvas(parent=root, image_queue=image_acquisition_thread.get_output_queue())

            print("Setting camera parameters...")
            camera.frames_per_trigger_zero_for_unlimited = 0
            camera.arm(2)
            camera.issue_software_trigger()

            print("Starting image acquisition thread...")
            image_acquisition_thread.start()

            print("App starting")
            root.mainloop()

            print("Waiting for image acquisition thread to finish...")
            image_acquisition_thread.stop()
            image_acquisition_thread.join()

            print("Closing resources...")

    print("App terminated. Goodbye!")



"""
Tiff Writing Example - tifffile

This example shows how to use Thorlabs TSI Cameras to write images to a disk using the tifffile library,
see https://pypi.org/project/tifffile/ for more information.

There are many TIFF-writing libraries for python, this example is meant to show how to integrate with tifffile.
The process should generally be the same with most TIFF-writing libraries, but results may vary.

In this example 10 images are going to be taken and saved to a single multipage TIFF file. The program will detect
if the camera has a color filter and will perform color processing if so.

One thing to note is that this program will save TIFFs in the camera's bit depth. Some image viewers may not recognize
this and will show the images as being much darker than expected. If you are experiencing dark images, we recommend
trying out various image viewers designed for scientific imaging such as ThorCam or ImageJ.

"""

try:
    # if on Windows, use the provided setup script to add the DLLs folder to the PATH
    from windows_setup import configure_path
    configure_path()
except ImportError:
    configure_path = None

import os
import tifffile

from thorlabs_tsi_sdk.tl_camera import TLCameraSDK
from thorlabs_tsi_sdk.tl_mono_to_color_processor import MonoToColorProcessorSDK
from thorlabs_tsi_sdk.tl_camera_enums import SENSOR_TYPE

NUMBER_OF_IMAGES = 10  # Number of TIFF images to be saved
OUTPUT_DIRECTORY = os.path.abspath(r'.')  # Directory the TIFFs will be saved to
FILENAME = 'image.tif'  # The filename of the TIFF

TAG_BITDEPTH = 32768
TAG_EXPOSURE = 32769

# delete image if it exists
if os.path.exists(OUTPUT_DIRECTORY + os.sep + FILENAME):
    os.remove(OUTPUT_DIRECTORY + os.sep + FILENAME)

with TLCameraSDK() as sdk:
    cameras = sdk.discover_available_cameras()
    if len(cameras) == 0:
        print("Error: no cameras detected!")

    with sdk.open_camera(cameras[0]) as camera:
        #  setup the camera for continuous acquisition
        camera.frames_per_trigger_zero_for_unlimited = 0
        camera.image_poll_timeout_ms = 2000  # 2 second timeout
        camera.arm(2)

        # save these values to place in our custom TIFF tags later
        bit_depth = camera.bit_depth
        exposure = camera.exposure_time_us

        # need to save the image width and height for color processing
        image_width = camera.image_width_pixels
        image_height = camera.image_height_pixels

        # initialize a mono to color processor if this is a color camera
        is_color_camera = (camera.camera_sensor_type == SENSOR_TYPE.BAYER)
        mono_to_color_sdk = None
        mono_to_color_processor = None
        if is_color_camera:
            mono_to_color_sdk = MonoToColorProcessorSDK()
            mono_to_color_processor = mono_to_color_sdk.create_mono_to_color_processor(
                camera.camera_sensor_type,
                camera.color_filter_array_phase,
                camera.get_color_correction_matrix(),
                camera.get_default_white_balance_matrix(),
                camera.bit_depth
            )

        # begin acquisition
        camera.issue_software_trigger()
        frames_counted = 0
        while frames_counted < NUMBER_OF_IMAGES:
            frame = camera.get_pending_frame_or_null()
            if frame is None:
                raise TimeoutError("Timeout was reached while polling for a frame, program will now exit")

            frames_counted += 1

            image_data = frame.image_buffer
            if is_color_camera:
                # transform the raw image data into RGB color data
                image_data = mono_to_color_processor.transform_to_48(image_data, image_width, image_height)
                image_data = image_data.reshape(image_height, image_width, 3)

            with tifffile.TiffWriter(OUTPUT_DIRECTORY + os.sep + FILENAME, append=True) as tiff:
                """
                    Setting append=True here means that calling tiff.save will add the image as a page to a multipage TIFF. 
                """
                tiff.save(data=image_data,  # np.ushort image data array from the camera
                          compress=0,   # amount of compression (0-9), by default it is uncompressed (0)
                          extratags=[(TAG_BITDEPTH, 'I', 1, bit_depth, False),  # custom TIFF tag for bit depth
                                     (TAG_EXPOSURE, 'I', 1, exposure, False)]  # custom TIFF tag for exposure
                          )
                """
                    If compress > 0 tifffile will compress the image using zlib - deflate compression.
                    Instead of an int a str can be supplied to specify a different compression algorithm;
                        e.g. compress = 'lzma'
                    View the tifffile source or online to see what is supported.
                """
                """
                    The extratags parameter allows the user to specify additional tags. Programs will typically ignore 
                    any tags from 32768 onward, which is where the bit depth and exposure have been placed. The 
                    syntax for extra tags is (tag_code, data_type_of_value, number_of_values, value, write_once).
                    View the tifffile source for more information.
                """
        camera.disarm()

        # we did not use context manager for color processor, so manually dispose of it
        if is_color_camera:
            try:
                mono_to_color_processor.dispose()
            except Exception as exception:
                print("Unable to dispose mono to color processor: " + str(exception))
            try:
                mono_to_color_sdk.dispose()
            except Exception as exception:
                print("Unable to dispose mono to color sdk: " + str(exception))

"""
Reading tiffs - to test that the tags from before worked, we're going to read back the tags on the first page. 
Note that custom TIFF tags are not going to be picked up by normal TIFF viewers, but can be read programmatically 
if the tag code is known.
"""
# open file

with tifffile.TiffFile(OUTPUT_DIRECTORY + os.sep + FILENAME) as tiff_read:
    if len(tiff_read.pages) < 1:
        raise ValueError("No pages were found in multipage TIFF")
    page_one = tiff_read.pages[0]
    print("First Image: Bit Depth = {} bpp, Exposure Time = {} ms".format(page_one.tags[str(TAG_BITDEPTH)].value,
                                                                          page_one.tags[str(TAG_EXPOSURE)].value/1000))



# -*- coding: utf-8 -*-
"""
Created on Tue Jul  5 16:27:13 2022

@author: AOG
"""



import cv2 as cv
#import pybundle
import numpy as np
import time
import sys
import os
import math
sys.path.append(os.path.abspath("..\\src\\widgets"))
sys.path.append(os.path.abspath("..\\src\\cameras"))
sys.path.append(os.path.abspath("..\\src\\threads"))

from BundleProcessor import BundleProcessor
from ImageAcquisitionThread import ImageAcquisitionThread

imageProcessor = BundleProcessor(10,10, mosaic = True)
imageProcessor.pyb.set_core_method(imageProcessor.pyb.FILTER)
imageProcessor.pyb.set_filter_size(1.4)
imageProcessor.pyb.set_bundle_loc((600,500,450))
imageProcessor.pyb.set_auto_mask(None)

rawImageBufferSize = 10

sourceFilename = (r"..\\..\\endomicroscope\\test\\data\\raw_example.tif")
imageThread = ImageAcquisitionThread('SimulatedCamera', rawImageBufferSize, filename=sourceFilename )
imageThread.get_camera().pre_load(100)
print("images pre-loaded")

imageThread.cam.set_frame_rate(100)
imageThread.start()
#imageProcessor.start()

nFrames = 100

t1 = time.perf_counter()
for i in range(1,nFrames):
    while imageThread.is_image_ready() is False:
        time.sleep(0.0001)
    a = imageThread.get_next_image()
    #print("n" + str(i))
    #print(imageThread.get_actual_fps())
    #print("Processed frame:" + str(i))
    # print(imageProcessor.is_image_ready())
    # imageProcessor.add_image(imgStack[i,:,:])
    # time.sleep(0.01)
    # currentProcessedImage = imageProcessor.get_latest_processed_image()
print((time.perf_counter() - t1) / nFrames)    
    
imageProcessor.stop() 
imageThread.stop()



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 28 18:17:10 2022

@author: AOG
"""
import context

from ImageAcquisitionThread import ImageAcquisitionThread

import matplotlib.pyplot as plt

import time


filename = r"test_data\stack_10.tif"

imThread = ImageAcquisitionThread('SimulatedCamera', 100, filename = filename)
imThread.get_camera().pre_load(10)

imThread.start()

try:
    while True:
        if imThread.is_image_ready():
            frame = imThread.get_next_image()
            print(frame)
                
except KeyboardInterrupt:
    pass


"""
Polling Example

This example shows how to open a camera, adjust some settings, and poll for images. It also shows how 'with' statements
can be used to automatically clean up camera and SDK resources.

"""

try:
    # if on Windows, use the provided setup script to add the DLLs folder to the PATH
    from windows_setup import configure_path
    configure_path()
except ImportError:
    configure_path = None

import numpy as np
from thorlabs_tsi_sdk.tl_camera import TLCameraSDK, OPERATION_MODE

NUM_FRAMES = 10  # adjust to the desired number of frames


with TLCameraSDK() as sdk:
    available_cameras = sdk.discover_available_cameras()
    if len(available_cameras) < 1:
        print("no cameras detected")

    with sdk.open_camera(available_cameras[0]) as camera:
        camera.exposure_time_us = 11000  # set exposure to 11 ms
        camera.frames_per_trigger_zero_for_unlimited = 0  # start camera in continuous mode
        camera.image_poll_timeout_ms = 1000  # 1 second polling timeout
        old_roi = camera.roi  # store the current roi
        """
        uncomment the line below to set a region of interest (ROI) on the camera
        """
        # camera.roi = (100, 100, 600, 600)  # set roi to be at origin point (100, 100) with a width & height of 500

        """
        uncomment the lines below to set the gain of the camera and read it back in decibels
        """
        #if camera.gain_range.max > 0:
        #    db_gain = 6.0
        #    gain_index = camera.convert_decibels_to_gain(db_gain)
        #    camera.gain = gain_index
        #    print(f"Set camera gain to {camera.convert_gain_to_decibels(camera.gain)}")

        camera.arm(2)

        camera.issue_software_trigger()

        for i in range(NUM_FRAMES):
            frame = camera.get_pending_frame_or_null()
            if frame is not None:
                print("frame #{} received!".format(frame.frame_count))

                frame.image_buffer  # .../ perform operations using the data from image_buffer

                #  NOTE: frame.image_buffer is a temporary memory buffer that may be overwritten during the next call
                #        to get_pending_frame_or_null. The following line makes a deep copy of the image data:
                image_buffer_copy = np.copy(frame.image_buffer)
            else:
                print("timeout reached during polling, program exiting...")
                break
        camera.disarm()
        camera.roi = old_roi  # reset the roi back to the original roi

#  Because we are using the 'with' statement context-manager, disposal has been taken care of.

print("program completed")



"""
Color Example

This example shows how to work with the mono to color processor module. It will open a camera and poll for an image,
but the focus will be on explaining how to color the image from the camera.

"""

try:
    # if on Windows, use the provided setup script to add the DLLs folder to the PATH
    from windows_setup import configure_path
    configure_path()
except ImportError:
    configure_path = None

from tl_camera import TLCameraSDK
from tl_mono_to_color_processor import MonoToColorProcessorSDK
from tl_mono_to_color_enums import COLOR_SPACE
from tl_color_enums import FORMAT

"""
    The MonoToColorProcessorSDK and MonoToColorProcessor objects can be used with context managers for automatic 
    clean up. This multi-context-manager 'with' statement opens both the camera sdk and the mono to color sdk at 
    once. 
"""
with TLCameraSDK() as camera_sdk, MonoToColorProcessorSDK() as mono_to_color_sdk:
    available_cameras = camera_sdk.discover_available_cameras()
    if len(available_cameras) < 1:
        raise ValueError("no cameras detected")

    with camera_sdk.open_camera(available_cameras[0]) as camera:
        camera.frames_per_trigger_zero_for_unlimited = 0  # start camera in continuous mode
        camera.image_poll_timeout_ms = 2000  # 2 second timeout
        camera.arm(2)

        """
            In a real-world scenario, we want to save the image width and height before color processing so that we 
            do not have to query it from the camera each time it is needed, which would slow down the process. It is 
            safe to save these after arming since the image width and height cannot change while the camera is armed.
        """
        image_width = camera.image_width_pixels
        image_height = camera.image_height_pixels

        camera.issue_software_trigger()

        frame = camera.get_pending_frame_or_null()
        if frame is not None:
            print("frame received!")
        else:
            raise ValueError("No frame arrived within the timeout!")

        camera.disarm()

        """
            When creating a mono to color processor, we want to initialize it using parameters from the camera.
        """
        with mono_to_color_sdk.create_mono_to_color_processor(
            camera.camera_sensor_type,
            camera.color_filter_array_phase,
            camera.get_color_correction_matrix(),
            camera.get_default_white_balance_matrix(),
            camera.bit_depth
        ) as mono_to_color_processor:
            """
                Once it is created, we can change the color space and output format properties. sRGB is the default 
                color space, and will usually give the best looking image. The output format will determine how the 
                transform image data will be structured.
            """
            mono_to_color_processor.color_space = COLOR_SPACE.SRGB  # sRGB color space
            mono_to_color_processor.output_format = FORMAT.RGB_PIXEL  # data is returned as sequential RGB values
            """
                We can also adjust the Red, Green, and Blue gains. These values amplify the intensity of their 
                corresponding colors in the transformed image. For example, if Blue and Green gains are set to 0 
                and the Red gain is 10, the resulting image will look entirely Red. The most common use case for these 
                properties will be for white balancing. By default they are set to model-specific values that gives 
                reasonably good white balance in typical lighting.
            """
            print("Red Gain = {red_gain}\nGreen Gain = {green_gain}\nBlue Gain = {blue_gain}\n".format(
                red_gain=mono_to_color_processor.red_gain,
                green_gain=mono_to_color_processor.green_gain,
                blue_gain=mono_to_color_processor.blue_gain
            ))
            """
                When we have all the settings we want for the mono to color processor, we call one of the transform_to 
                functions to get a color image. 
            """
            # this will give us a resulting image with 3 channels (RGB) and 16 bits per channel, resulting in 48 bpp
            color_image_48_bpp = mono_to_color_processor.transform_to_48(frame.image_buffer, image_width, image_height)

            # this will give us a resulting image with 4 channels (RGBA) and 8 bits per channel, resulting in 32 bpp
            color_image_32_bpp = mono_to_color_processor.transform_to_32(frame.image_buffer, image_width, image_height)

            # this will give us a resulting image with 3 channels (RGB) and 8 bits per channel, resulting in 24 bpp
            color_image_24_bpp = mono_to_color_processor.transform_to_24(frame.image_buffer, image_width, image_height)

            # from here, perform any actions you need to using the color image

#  Because we are using the 'with' statement context-manager, disposal has been taken care of.

print("program completed")



"""
tl_color_enums.py
"""

from enum import IntEnum


class __CTypesEnum(IntEnum):
    @classmethod
    def from_param(cls, obj):
        return int(obj)


class FILTER_ARRAY_PHASE(__CTypesEnum):
    """
    The FILTER_ARRAY_PHASE enumeration lists all the possible values that a pixel in a Bayer pattern color arrangement 
    could assume.

    The classic Bayer pattern is::

        -----------------------
        |          |          |
        |    R     |    GR    |
        |          |          |
        -----------------------
        |          |          |
        |    GB    |    B     |
        |          |          |
        -----------------------

    where:
    
    - R = a red pixel
    - GR = a green pixel next to a red pixel
    - B = a blue pixel
    - GB = a green pixel next to a blue pixel
   
    The primitive pattern shown above represents the fundamental color pixel arrangement in a Bayer pattern
    color sensor.  The basic pattern would extend in the X and Y directions in a real color sensor containing
    millions of pixels.
   
    Notice that the color of the origin (0, 0) pixel logically determines the color of every other pixel.
   
    It is for this reason that the color of this origin pixel is termed the color "phase" because it represents
    the reference point for the color determination of all other pixels.
   
    Every TSI color camera provides the sensor specific color phase of the full frame origin pixel as a discoverable
    parameter.

    """
    BAYER_RED = 0
    """
    A red pixel.
    
    """
    BAYER_BLUE = 1
    """
    A blue pixel.
    
    """
    GREEN_LEFT_OF_RED = 2
    """
    A green pixel next to a red pixel.
    
    """
    GREEN_LEFT_OF_BLUE = 3
    """
    A green pixel next to a blue pixel.
    
    """


class FORMAT(__CTypesEnum):
    """
    The FORMAT enumeration lists all the possible options for specifying the order of
    color pixels in input and/or output buffers.
   
    Depending on the context, it can specify:

    - the desired pixel order that a module must use when writing color pixel data into an output buffer
    - the pixel order that a module must use to interpret data in an input buffer.
    
    """
    BGR_PLANAR = 0
    """
    The color pixels blue, green, and red are grouped in separate planes in the buffer:\n
    BBBBB... GGGGG... RRRRR....
    
    """
    BGR_PIXEL = 1
    """
    The color pixels blue, green, and red are clustered and stored consecutively in the following pattern:\n 
    BGRBGRBGR...
    
    """
    RGB_PIXEL = 2
    """
    The color pixels blue, green, and red are clustered and stored consecutively in the following pattern:\n
    RGBRGBRGB...
    
    """


class FILTER_TYPE(__CTypesEnum):
    """
    The FILTER_TYPE enumeration lists all the possible filter options for color cameras
    
    """
    BAYER = 0
    """
    A Bayer pattern color sensor.
    
    """



"""
tl_polarization_enums.py
"""

from enum import IntEnum


class _CTypesEnum(IntEnum):
    @classmethod
    def from_param(cls, obj):
        return int(obj)


class POLAR_PHASE(_CTypesEnum):
    """
    The possible polarization angle values (in degrees) for a pixel in a polarization sensor. The polarization phase
    pattern of the sensor is::

        -------------
        | + 0 | -45 |
        -------------
        | +45 | +90 |
        -------------

    The primitive pattern shown above represents the fundamental polarization phase arrangement in a polarization
    sensor. The basic pattern would extend in the X and Y directions in a real polarization sensor containing millions
    of pixels. Notice that the phase of the origin (0, 0) pixel logically determines the phase of every other pixel.
    It is for this reason that the phase of this origin pixel is termed the polarization "phase" because it represents
    the reference point for the phase determination of all other pixels.


    """
    PolarPhase0 = 0
    """
    0 degrees polarization phase 
    
    """
    PolarPhase45 = 1
    """
    45 degrees polarization phase 

    """
    PolarPhase90 = 2
    """
    90 degrees polarization phase 

    """
    PolarPhase135 = 3
    """
    135 (-45) degrees polarization phase 

    """



"""
tl_camera_enums.py
"""

from enum import IntEnum


class _CTypesEnum(IntEnum):
    @classmethod
    def from_param(cls, obj):
        return int(obj)


class OPERATION_MODE(_CTypesEnum):
    """
    The OPERATION_MODE enumeration defines the available modes for a camera.

    """
    SOFTWARE_TRIGGERED = 0
    """
    Use software operation mode to generate one or more frames per trigger or to run continuous video mode.
    
    """
    HARDWARE_TRIGGERED = 1
    """
    Use hardware triggering to generate one or more frames per trigger by issuing hardware signals.
    
    """
    BULB = 2
    """
    Use bulb-mode triggering to generate one or more frames per trigger by issuing hardware signals. Please refer to 
    the camera manual for signaling details.
    
    """
    RESERVED1 = 3  # Reserved for internal use.
    RESERVED2 = 4  # Reserved for internal use.


class SENSOR_TYPE(_CTypesEnum):
    """
    This describes the physical capabilities of the camera sensor.

    """
    MONOCHROME = 0
    """
    Each pixel of the sensor indicates an intensity.
    
    """
    BAYER = 1
    """
    The sensor has a bayer-patterned filter overlaying it, allowing the camera SDK to distinguish red, green, and blue 
    values.
    
    """
    MONOCHROME_POLARIZED = 2
    """
     The sensor has a polarization filter overlaying it allowing the camera to capture polarization information from 
     the incoming light.

    """


class TRIGGER_POLARITY(_CTypesEnum):
    """
    The TRIGGER_POLARITY enumeration defines the options available for specifying the hardware trigger polarity. These
    values specify which edge of the input trigger pulse that will initiate image acquisition.

    """
    ACTIVE_HIGH = 0
    """
    Acquire an image on the RISING edge of the trigger pulse.
    
    """
    ACTIVE_LOW = 1
    """
    Acquire an image on the FALLING edge of the trigger pulse.

    """


class EEP_STATUS(_CTypesEnum):
    """
    The EEP_STATUS enumeration defines the options available for specifying the device's EEP mode. Equal Exposure Pulse
    (EEP) mode is an LVTTL-level signal that is active during the time when all rows have been reset during rolling
    reset, and the end of the exposure time (and the beginning of rolling readout). The signal can be used to control
    an external light source that will be on only during the equal exposure period, providing the same amount of
    exposure for all pixels in the ROI. EEP mode can be enabled, but may be active or inactive depending on the
    current exposure value. If EEP is enabled in bulb mode, it will always give a status of Bulb.

    """
    DISABLED = 0
    """
    EEP mode is disabled.
    
    """
    ENABLED_ACTIVE = 1
    """
    EEP mode is enabled and currently active.
    
    """
    ENABLED_INACTIVE = 2
    """
    EEP mode is enabled, but due to an unsupported exposure value, currently inactive.
    
    """
    ENABLED_BULB = 3
    """
    EEP mode is enabled in bulb mode.
    
    """


class DATA_RATE(_CTypesEnum):
    """
    The DATA_RATE enumeration defines the options for setting the desired image data delivery rate.

    """
    RESERVED1 = 0  # A RESERVED value (DO NOT USE).
    RESERVED2 = 1  # A RESERVED value (DO NOT USE).
    FPS_30 = 2
    """
    Sets the device to deliver images at 30 frames per second.
    
    """
    FPS_50 = 3
    """
    Sets the device to deliver images at 50 frames per second.
    
    """


class USB_PORT_TYPE(_CTypesEnum):
    """
    The USB_PORT_TYPE enumeration defines the values the SDK uses for specifying the USB bus speed. These values are
    returned by SDK API functions and callbacks based on the type of physical USB port that the device is connected to.

    """
    USB1_0 = 0
    """
    The device is connected to a USB 1.0/1.1 port (1.5 Mbits/sec or 12 Mbits/sec).
    
    """
    USB2_0 = 1
    """
    The device is connected to a USB 2.0 port (480 Mbits/sec).
    
    """
    USB3_0 = 2
    """
    The device is connected to a USB 3.0 port (5000 Mbits/sec).
    
    """


class TAPS(_CTypesEnum):
    """
    Scientific CCD cameras support one or more taps. After exposure is complete, a CCD pixel array holds the charges
    corresponding to the amount of light collected at beach pixel location. The data is then read out through 1, 2,
    or 4 channels at a time.

    """
    SINGLE_TAP = 0
    """
    Charges are read out through a single analog-to-digital converter.
    
    """
    DUAL_TAP = 1
    """
    Charges are read out through two analog-to-digital converters.
    
    """
    QUAD_TAP = 2
    """
    Charges are read out through four analog-to-digital converters.
    
    """


class COMMUNICATION_INTERFACE(_CTypesEnum):
    """
    Used to identify what interface the camera is currently using. If using USB, the specific USB version can also be
    identified using USB_PORT_TYPE.

    """
    GIG_E = 0
    """
    The camera uses the GigE Vision (GigE) interface standard.
    
    """
    LINK = 1
    """
    The camera uses the CameraLink serial-communication-protocol standard.
    
    """
    USB = 2
    """
    The camera uses a USB interface.
    
    """



"""
Polarization Example

This example shows how to work with the polarization processor module. It will open a camera and poll for an image,
but the focus will be on explaining how to process the image from the camera. At the end, 4 consecutive images should
pop up showing intensity data, azimuth data, degree of linear polarization (DoLP) data, and a Quad View.

"""

try:
    # if on Windows, use the provided setup script to add the DLLs folder to the PATH
    from windows_setup import configure_path
    configure_path()
except ImportError:
    configure_path = None

import numpy as np
from PIL import Image

from thorlabs_tsi_sdk.tl_camera import TLCameraSDK
from thorlabs_tsi_sdk.tl_camera_enums import SENSOR_TYPE
from thorlabs_tsi_sdk.tl_polarization_processor import PolarizationProcessorSDK

"""
    The PolarizationProcessorSDK and PolarizationProcessor objects can be used with context managers for automatic 
    clean up. This multi-context-manager 'with' statement opens both the camera sdk and the polarization sdk at 
    once. 
"""
with TLCameraSDK() as camera_sdk, PolarizationProcessorSDK() as polarization_sdk:
    available_cameras = camera_sdk.discover_available_cameras()
    if len(available_cameras) < 1:
        raise ValueError("no cameras detected")

    with camera_sdk.open_camera(available_cameras[0]) as camera:
        camera.frames_per_trigger_zero_for_unlimited = 0  # start camera in continuous mode
        camera.image_poll_timeout_ms = 2000  # 2 second timeout
        camera.arm(2)

        """
            In a real-world scenario, we want to save the image width and height before color processing so that we 
            do not have to query it from the camera each time it is needed, which would slow down the process. It is 
            safe to save these after arming since the image width and height cannot change while the camera is armed.
        """
        image_width = camera.image_width_pixels
        image_height = camera.image_height_pixels

        camera.issue_software_trigger()

        frame = camera.get_pending_frame_or_null()
        if frame is not None:
            print("frame received!")
        else:
            raise ValueError("No frame arrived within the timeout!")

        camera.disarm()

        if camera.camera_sensor_type is not SENSOR_TYPE.MONOCHROME_POLARIZED:
            raise ValueError("Polarization processing should only be done with polarized cameras")

        camera_polar_phase = camera.polar_phase
        camera_bit_depth = camera.bit_depth
        """
        We're scaling to 8-bits in this example so we can easily convert to PIL Image objects.
        """
        max_output_value = 255

        with polarization_sdk.create_polarization_processor() as polarization_processor:
            """
            Convert the raw sensor data to polarization image data. We will convert to each of the available outputs: 
            intensity, azimuth, and degree of linear polarization (DoLP).
            """
            output_intensity = polarization_processor.transform_to_intensity(camera_polar_phase,
                                                                             frame.image_buffer,
                                                                             0,  # origin x
                                                                             0,  # origin y
                                                                             image_width,
                                                                             image_height,
                                                                             camera_bit_depth,
                                                                             max_output_value)
            output_azimuth = polarization_processor.transform_to_azimuth(camera_polar_phase,
                                                                         frame.image_buffer,
                                                                         0,  # origin x
                                                                         0,  # origin y
                                                                         image_width,
                                                                         image_height,
                                                                         camera_bit_depth,
                                                                         max_output_value)
            output_dolp = polarization_processor.transform_to_dolp(camera_polar_phase,
                                                                   frame.image_buffer,
                                                                   0,  # origin x
                                                                   0,  # origin y
                                                                   image_width,
                                                                   image_height,
                                                                   camera_bit_depth,
                                                                   max_output_value)
            """
            Reshape the 1D output arrays to be 2D arrays (image_width x image_height)
            """
            output_intensity = output_intensity.reshape(image_height, image_width)
            output_azimuth = output_azimuth.reshape(image_height, image_width)
            output_dolp = output_dolp.reshape(image_height, image_width)
            """
            Convert the polarization image data to PIL Image types
            """
            intensity_image = Image.fromarray(output_intensity)
            azimuth_image = Image.fromarray(output_azimuth)
            dolp_image = Image.fromarray(output_dolp)
            """
            We're going to display each of the images created using PIL's show() method. This tries to open the images 
            using your default image viewer, and may fail on some configurations.
            """
            intensity_image.show()
            azimuth_image.show()
            dolp_image.show()
            """
            Lastly, we'll construct a QuadView image that is useful for visualizing each polar rotation: 0, 45, 90, and 
            -45 degrees. The sensor on the polarized camera has a filter in front of it that is composed of 2x2 pixel 
            sections that look like the following pattern: 

            -------------
            | +45 | +90 |
            -------------
            | + 0 | -45 |
            -------------

            It is always 2x2, but the ordering of the rotations may differ depending on your camera model. The top left 
            rotation (the 'origin' rotation) is always equal to the camera_polar_phase that was queried earlier. We'll 
            use array splicing to extract each of the rotations and separate them visually. If you are familiar with 
            manipulating color image arrays, this is similar to pulling out the R, G, and B components of an RGB image.
            """
            unprocessed_image = frame.image_buffer.reshape(image_height, image_width)  # this is the raw image data
            unprocessed_image = unprocessed_image >> camera_bit_depth - 8  # scale to 8 bits for easier displaying
            output_quadview = np.zeros(shape=(image_height, image_width))  # initialize array for QuadView data
            # Top Left Quadrant =
            output_quadview[0:int(image_height / 2), 0:int(image_width / 2)] = \
                unprocessed_image[0::2, 0::2]  # (0,0): top left rotation == camera_polar_phase
            # Top Right Quadrant =
            output_quadview[0:int(image_height / 2), int(image_width / 2):image_width] = \
                unprocessed_image[0::2, 1::2]  # (0,1): top right rotation
            # Bottom Left Quadrant =
            output_quadview[int(image_height / 2):image_height, 0:int(image_width / 2)] = \
                unprocessed_image[1::2, 0::2]  # (1,0): bottom left rotation
            # Bottom Right Quadrant =
            output_quadview[int(image_height / 2):image_height, int(image_width / 2):image_width] = \
                unprocessed_image[1::2, 1::2]  # (1,1): bottom right rotation
            # Display QuadView
            quadview_image = Image.fromarray(output_quadview)
            quadview_image.show()

#  Because we are using the 'with' statement context-manager, disposal has been taken care of.

print("program completed")
